/*
    (C) 1999-2000 AROS - The Amiga Research OS
    $Id: join.c,v 1.2 2010/10/04 16:36:54 piru Exp $
 
    Desc: Join - Create a single file from several.
    Lang: English
*/

/******************************************************************************
 
 
    NAME
 
    Join [FILE] {(file | pattern)} AS|TO (filename)
 
    SYNOPSIS
 
    FILE/M/A,AS=TO/K/A
 
    LOCATION
 
    Workbench:C
 
    FUNCTION
 
    Join makes one big file of all listed files by putting them together
    in the order given. The destination file may not have the same name
    as any of input files. You must supply a destination file name. The
    original files remian unchanged. Any number of files can be Join:ed in
    one operation.
 
    INPUTS
 
    FILE   --  files to join
    TO=AS  --  the name of the combined file
 
    RESULT
 
    NOTES
 
    EXAMPLE
 
    BUGS
 
    SEE ALSO
 
    INTERNALS
 
    HISTORY
 
    30.11.2000  SDuvan  added pattern matching support
 
******************************************************************************/

#define  DEBUG  0
#define USE_SKIPDIRS   1   /* skip directories quietly */

#include <exec/devices.h>
#include <exec/io.h>
#include <exec/semaphores.h>
#include <exec/types.h>
#include <exec/memory.h>
#include <dos/dos.h>
#include <dos/dosasl.h>
#include <dos/exall.h>

#include <aros/debug.h>

#include <proto/exec.h>
#include <proto/dos.h>

#include <stdio.h>

#define ARG_TEMPLATE "FILE/M/A,AS=TO/K/A"
#define ARG_FILE     0
#define ARG_AS       1
#define ARG_COUNT    2

//#define BUFFERSIZE   32768    /* Buffersize used when reading & writing */
#define BUFFERSIZE   262144   /* Buffersize used when reading & writing */

/****** StringID's for getstring() **************************************/

#define STR_ABORTED       1   /* Aborted                                */
#define STR_REMOVINGDEST  2   /* Removing destination-file              */
#define STR_ERR_OPENREAD  3   /* Error opening file for reading         */
#define STR_ERR_OPENWRITE 4   /* Error opening file for writing         */
#define STR_ERR_NOMEM     5   /* Error allocating memory                */
#define STR_ERR_READING   6   /* Error while reading from file          */
#define STR_ERR_WRITING   7   /* Error while writing to file            */

/****** Version- and error-header ***************************************/

#include "join_version.h"

static const char version[] = VERSTAG;

static const char ERROR_HEADER[] __attribute__((section(".rodata"))) = "Join";

/****** Prototypes for local functions **********************************/

static LONG append(BPTR destfile, STRPTR srcfilename, struct DosLibrary *DOSBase, struct ExecBase *SysBase);
static STRPTR getstring(LONG stringid );
static int doJoin(STRPTR *files, BPTR destfile, struct DosLibrary *DOSBase, struct ExecBase *SysBase);

/****** Functions *******************************************************/


int main( void )
{
    struct ExecBase	*SysBase;
    struct DosLibrary	*DOSBase;
    struct RDArgs	*rda = NULL ;

    IPTR    args[ARG_COUNT] = { NULL , NULL };
    STRPTR *files;
    STRPTR  destination;
    BPTR    destfile = NULL;
    LONG    rc = RETURN_FAIL;

    SysBase=*((struct ExecBase**) 4);
    if ( (DOSBase = (struct DosLibrary *)OpenLibrary("dos.library",37)) )
    {
        if( (rda = ReadArgs( ARG_TEMPLATE , args , NULL )) )
        {
            /*if( args[ARG_FILE] && args[ARG_AS] )
            {*/
                destination = (STRPTR)args[ARG_AS];
                files = (STRPTR *)args[ARG_FILE];

                if( (destfile = Open( destination , MODE_NEWFILE )) )
                {
                    rc = doJoin(files, destfile, DOSBase, SysBase);

                    Close(destfile);

                    if (rc != RETURN_OK)
                    {
                        Printf(", %s.\n" , (ULONG)getstring(STR_REMOVINGDEST));
                        DeleteFile(destination);
                    }
                }
                else
                {
                    PrintFault(IoErr(), NULL);
                    Printf("%s: %s: '%s'\n", (ULONG)ERROR_HEADER,
                           (ULONG)getstring(STR_ERR_OPENWRITE), (ULONG)destination);
                }
            /*}*/

            FreeArgs(rda);
        }
        else
        {
            PrintFault(IoErr(), ERROR_HEADER);
        }

        CloseLibrary((struct Library *)DOSBase);
    }

    return rc;
}


#define  MAX_PATH_LEN  512


static
int doJoin(STRPTR *files, BPTR destfile, struct DosLibrary *DOSBase, struct ExecBase *SysBase)
{
    struct AnchorPath *ap;

    LONG  i;			/* Loop variable over patterns */
    LONG  match;		/* Loop variable over files */
    LONG  rc = RETURN_OK;

    ap = (struct AnchorPath *)AllocVec(sizeof(struct AnchorPath) +
                                       MAX_PATH_LEN, MEMF_CLEAR);

    if (ap == NULL)
    {
        SetIoErr(ERROR_NO_FREE_STORE);
        return RETURN_FAIL;
    }

    ap->ap_Strlen = MAX_PATH_LEN;

    /* Loop over the arguments */
    for (i = 0; files[i] != NULL; i++)
    {
        BOOL flag = FALSE;

        ap->ap_BreakBits  = SIGBREAKF_CTRL_C;
        ap->ap_FoundBreak = 0;

        for (match = MatchFirst(files[i], ap);
             match == 0;
             match = MatchNext(ap))
        {
            flag = TRUE;

            #if USE_SKIPDIRS
            if (ap->ap_Info.fib_DirEntryType > 0 && ap->ap_Info.fib_DirEntryType != ST_SOFTLINK)
            {
                continue;
            }
            #endif /* USE_SKIPDIRS */

            if(append(destfile, ap->ap_Buf, DOSBase, SysBase) != RETURN_OK )
            {
                Printf("%s: %s", (ULONG)ERROR_HEADER, (ULONG)getstring(STR_ABORTED));
                rc = RETURN_FAIL;
                break;
            }
        }

        if (ap->ap_FoundBreak & SIGBREAKF_CTRL_C)
        {
            PrintFault(IoErr(), NULL);
            SetIoErr(ERROR_BREAK);
            rc = RETURN_FAIL;
            break;
        }

        /* Try directly if it's not a pattern, NOTE: Re-uses ap->ap_Buf - Piru */
        if (!flag && ParsePattern(files[i], (STRPTR) ap->ap_Buf, MAX_PATH_LEN) != 1)
        {
            if(append(destfile, files[i], DOSBase, SysBase) != RETURN_OK )
            {
                Printf("%s: %s", (ULONG)ERROR_HEADER, (ULONG)getstring(STR_ABORTED));
                rc = RETURN_FAIL;
                break;
            }
        }
    }

    FreeVec(ap);

    return rc;
}


static LONG append(BPTR destfile, STRPTR srcfilename, struct DosLibrary *DOSBase, struct ExecBase *SysBase)
{
    BYTE  *buffer       = NULL;
    LONG   actualLength = 0;
    BPTR   srcfile      = NULL;

    BOOL   rc           = RETURN_OK;

    if ( (buffer = AllocMem( BUFFERSIZE , MEMF_ANY )) )
    {
        if ( (srcfile = Open( srcfilename , MODE_OLDFILE )) )
        {
            ULONG brk;

            while (!(brk = SetSignal(0,0) & SIGBREAKF_CTRL_C) &&
                   (actualLength = Read(srcfile, buffer, BUFFERSIZE)) != -1)
            {
                if (Write(destfile, buffer, actualLength) != actualLength)
                {
                    PrintFault(IoErr(), NULL);
                    Printf( "%s: %s.\n", (ULONG)ERROR_HEADER,
                            (ULONG)getstring(STR_ERR_WRITING));
                    rc = RETURN_FAIL;

                    break;
                }

                if (actualLength < BUFFERSIZE)
                {
                    break;
                }
            }

            if (actualLength == -1)
            {
                PrintFault(IoErr(), NULL);
                Printf( "%s: %s.\n", (ULONG)ERROR_HEADER,
                       (ULONG)getstring(STR_ERR_READING));
                rc = RETURN_FAIL;
            }

            if (brk)
            {
                PrintFault(ERROR_BREAK, NULL);
                SetIoErr(ERROR_BREAK);
                rc = RETURN_FAIL;
            }

            Close(srcfile);
        }
        else
        {
            LONG ioerr = IoErr();
            #if USE_SKIPDIRS
            if (ioerr != ERROR_OBJECT_WRONG_TYPE)
            #endif /* USE_SKIPDIRS */
            {
                PrintFault(ioerr, NULL);
                Printf("%s: %s: '%s'\n", (ULONG)ERROR_HEADER,
                       (ULONG)getstring(STR_ERR_OPENREAD), (ULONG)srcfilename);

                rc = RETURN_FAIL;
            }
        }

        FreeMem(buffer, BUFFERSIZE);
    }
    else
    {
        PrintFault(IoErr(), NULL);
        Printf("%s: %s.\n", (ULONG)ERROR_HEADER, (ULONG)getstring(STR_ERR_NOMEM));
        rc = RETURN_FAIL;
    }

    return rc;
}


STRPTR getstring(LONG stringid)
{
    switch(stringid)
    {
    case STR_ABORTED:
        return "Aborted";

    case STR_REMOVINGDEST:
        return "removed incomplete destination-file";

    case STR_ERR_OPENREAD:
        return "Could not open file for reading";

    case STR_ERR_OPENWRITE:
        return "Could not open file for writing";

    case STR_ERR_NOMEM:
        return "Could not allocate memory";

    case STR_ERR_READING:
        return "Error while reading";

    case STR_ERR_WRITING:
        return "Error while writing";

    default:
        return "[Error: Unknown StringID]";
    }
}
